import pandas as pd

CAR_NAMES_BY_REPLAY_NAME = {
    "standard-kart": "Standard Kart",
}


class Driver:
    def __init__(self, driver_id: str, driver_replay_dict: dict) -> None:
        self.assign_properties(driver_id, driver_replay_dict)
        self.series = None

    def assign_properties(self, did: str, d: dict) -> None:
        self.id = did

        self.name = d["racerName"]
        self.platform = d["platform"]
        self.is_ai = d["isAITeam"]

        self.colors = d["driverSkinLivery"][1]
        self.vehicle_colors = d["vehicleLivery"][1]

        self.car = CAR_NAMES_BY_REPLAY_NAME[d["vehicle"]]

    def as_series(self) -> pd.Series:
        if self.series is None:
            property_attributes = [
                attr
                for attr in dir(self)
                if not attr.startswith("__")
                and not callable(getattr(self, attr))
                and attr not in ["series", "id"]
            ]

            indices = property_attributes

            data = [getattr(self, attr) for attr in property_attributes]

            self.series = pd.Series(data=data, index=indices, name=self.id)

        return self.series

    def __str__(self) -> str:
        return self.name



from ksup_analyzer.event.Result import RaceResultsDataFrame

import matplotlib.pyplot as plt
import numpy as np


class LapPositionTablePlot:
    def __init__(self, df: RaceResultsDataFrame, race) -> None:
        self.legend = df["name"]
        self.lap_position = df.lap_position_table
        self.race = race

    def plot(self, fileprefix=None):
        fig = plt.figure(layout="constrained")
        fig.set_size_inches(20, 10)  # size of figure
        plt.rc("grid", linestyle=":", color="black")  # style of grid

        ax = plt.subplot(111)
        ax.plot(self.lap_position, linewidth=2)
        ax.set_xlabel("Laps", fontsize=20)
        ax.set_ylabel("Position", fontsize=20)
        ax.grid(True)

        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
        ax.legend(self.legend, loc="center left", bbox_to_anchor=(1, 0.5), fontsize=16)

        x_labels = ["Start"]
        for lap in range(len(self.lap_position) - 1):
            x_labels.append(f"Lap {lap+1}")

        # Show all ticks and label them with the respective list entries
        ax.set_xticks(np.arange(len(x_labels)), labels=x_labels)
        ax.set_yticks(np.arange(0, 12))
        ax.set_ylim(12, 0)

        plt.setp(ax.get_xticklabels(), rotation=90)
        # plt.show()
        dt = self.race.datetime_race
        if fileprefix == None:
            fig.savefig(
                f"export//{dt.strftime('%d_%m_%Y_%H_%M_%S')}_LapPosition.png", dpi=150
            )
        else:
            fig.savefig(f"export//{fileprefix}_LapPosition.png", dpi=150)


class GapToWinnerTablePlot:
    def __init__(self, df: RaceResultsDataFrame, race) -> None:
        self.legend = df["name"]
        self.gap_table = df.gap_to_winner_table
        self.race = race

    def plot(self, ymin=-15, ymax=15, fileprefix=None):
        fig = plt.figure(layout="constrained")
        fig.set_size_inches(20, 10)  # size of figure
        plt.rc("grid", linestyle=":", color="black")  # style of grid

        ax = plt.subplot(111)
        ax.plot(self.gap_table, linewidth=2)
        ax.set_xlabel("Laps", fontsize=20)
        ax.set_ylabel("Gap to winner (s)", fontsize=20)
        ax.grid(True)

        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
        ax.legend(self.legend, loc="center left", bbox_to_anchor=(1, 0.5), fontsize=16)

        x_labels = ["Start"]
        for lap in range(len(self.gap_table) - 1):
            x_labels.append(f"Lap {lap+1}")

        # Show all ticks and label them with the respective list entries
        ax.set_xticks(np.arange(len(x_labels)), labels=x_labels)
        ax.set_yticks(np.arange(ymin, ymax))
        ax.set_ylim(ymax, ymin)

        plt.setp(ax.get_xticklabels(), rotation=90)
        dt = self.race.datetime_race
        if fileprefix == None:
            fig.savefig(
                f"export//{dt.strftime('%d_%m_%Y_%H_%M_%S')}_GapToWinner.png", dpi=150
            )
        else:
            fig.savefig(f"export//{fileprefix}_GapToWinner.png", dpi=150)


class GapToLeaderTablePlot:
    def __init__(self, df: RaceResultsDataFrame, race) -> None:
        self.legend = df["name"]
        self.gap_table = df.gap_to_leader_table
        self.race = race

    def plot(self, ymin=-1, ymax=30, fileprefix=None):
        fig = plt.figure(layout="constrained")
        fig.set_size_inches(20, 10)  # size of figure
        plt.rc("grid", linestyle=":", color="black")  # style of grid

        ax = plt.subplot(111)
        ax.plot(self.gap_table, linewidth=2)
        ax.set_xlabel("Laps", fontsize=20)
        ax.set_ylabel("Gap to leader (s)", fontsize=20)
        ax.grid(True)

        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
        ax.legend(self.legend, loc="center left", bbox_to_anchor=(1, 0.5), fontsize=16)

        x_labels = ["Start"]
        for lap in range(len(self.gap_table) - 1):
            x_labels.append(f"Lap {lap+1}")

        # Show all ticks and label them with the respective list entries
        ax.set_xticks(np.arange(len(x_labels)), labels=x_labels)
        ax.set_yticks(np.arange(ymin, ymax))
        ax.set_ylim(ymax, ymin)

        plt.setp(ax.get_xticklabels(), rotation=90)
        dt = self.race.datetime_race
        if fileprefix == None:
            fig.savefig(
                f"export//{dt.strftime('%d_%m_%Y_%H_%M_%S')}_GapToLeader.png", dpi=150
            )
        else:
            fig.savefig(f"export//{fileprefix}_GapToLeader.png", dpi=150)



import pandas as pd
from datetime import datetime
from ksup_analyzer.event.Result import RaceResult, QualiResult, RaceResultsDataFrame

TRACK_NAMES_BY_REPLAY_NAMES = {
    "sunny-side-track": {"name": "Sunny Side Park", "layouts": {"Path-A": "GP"}},
    "sugar-hill-track": {
        "name": "Sugar Hill",
        "layouts": {"Default-Path": "Full", "Reverse-Path": "Full – Reverse"},
    },
    "maple-ridge-track": {
        "name": "Maple Ridge",
        "layouts": {"Path-A": "Full", "Reverse": "Full – Reverse"},
    },
    "rennvoort-track": {
        "name": "Rennvoort",
        "layouts": {"Path-A": "Full", "Reverse": "Full – Reverse"},
    },
    "rivazza-track": {"name": "Faenza", "layouts": {"Path-A": "GP"}},
    "magdalena-track": {
        "name": "Magdalena",
        "layouts": {"Path-A": "Full", "Path-B": "Club"},
    },
    "copperwood-track": {
        "name": "Cooperwood",
        "layouts": {
            "Path-A": "Full",
            "Path-B": "Club",
            "Full-Reverse": "Full – Reverse",
            "Club-Reverse": "Club – Reverse",
        },
    },
    "siena-track": {
        "name": "Siena",
        "layouts": {"Path-A": "Full", "Reverse": "Reverse"},
    },
    "whistlevalley-track": {
        "name": "Whistle Valley",
        "layouts": {"Path-A": "Full", "Reverse": "Full – Reverse"},
    },
    "hidden-lake-track": {
        "name": "Hidden Lake",
        "layouts": {
            "Path-A": "Standard",
            "Path-B": "Reverse",
        },
    },
    "centrifuge-track": {
        "name": "Centrifuge",
        "layouts": {
            "Path-A": "Standard",
            "Path-B": "Reverse",
            "Path-C": "Club",
            "Path-D": "Club - Reverse",
        },
    },
    "tilksport-track": {
        "name": "Tilksport",
        "layouts": {
            "Path-A": "Full",
            "Path-B": "Club",
            "Path-C": "Rallycross",
            "Full-Reverse": "Full – Reverse",
            "Club-Reverse": "Club – Reverse",
        },
    },
    "thunder-valley-track": {
        "name": "Thunder Point",
        "layouts": {
            "Path-A": "Full",
            "Path-B": "Club",
            "Path-A-Reverse": "Full – Reverse",
            "Path-B-Reverse": "Club – Reverse",
        },
    },
    "oval1-track": {"name": "Bullseye Speedwey", "layouts": {"Path-A": "Oval"}},
    "speedopolis-track": {"name": "Speedopolis", "layouts": {"Path-A": "Oval"}},
    "dirt-oval-track": {"name": "Poke Bowl", "layouts": {"Path-A": "Oval"}},
    "grand-canyon-track": {
        "name": "Interstate",
        "layouts": {"Path-A": "Rallycross", "Reverse": "Rallycross – Reverse"},
    },
    "buffalo-track": {
        "name": "Buffalo Hill",
        "layouts": {"Path-A": "Rallycross", "Path-B": "Club", "Buffalo Hill": "Mini"},
    },
    "lost-lagoons-track": {"name": "Lost Lagoons", "layouts": {"Path-A": "Rallycross"}},
}


class Event:
    def __init__(self, lineup, race, quali=None) -> None:
        self.lineup = lineup
        self.race = race
        self.quali = quali

    def create_result_dataframe(self) -> None:
        # The main result dataframe will be a subclass of pd.DataFrame.
        # This allows us to create custom methods on it.
        race_result_df = RaceResultsDataFrame(
            [result.as_series() for result in self.race.results]
        )

        if self.quali:
            quali_df = pd.concat(
                [result.as_series() for result in self.quali.results], axis=1
            ).transpose()

            race_result_df = race_result_df.join(
                quali_df, lsuffix="_race", rsuffix="_quali"
            )
        else:
            # even if we do not have quali data, we want the race columns to be suffixed in the same way
            race_result_df.columns = [col + "_race" for col in race_result_df.columns]

        lineup_df = pd.concat(
            [driver.as_series() for driver in self.lineup.drivers], axis=1
        ).transpose()

        self.__result_df = race_result_df.join(lineup_df)

    @property
    def result_df(self) -> RaceResultsDataFrame:
        return self.__result_df

    def run_result_calculations(self) -> None:
        self.__result_df._run_result_calculations()


class Session:
    def __init__(self, race_file_content: dict) -> None:
        self.assign_properties(race_file_content)

    def assign_properties(self, r: dict) -> None:
        self.datetime_utc = datetime.strptime(r["timeStampUtc"], "%Y%m%dT%H:%M:%SZ")

        # location = track without layout
        self.location = TRACK_NAMES_BY_REPLAY_NAMES[r["track"]]["name"]
        self.layout = TRACK_NAMES_BY_REPLAY_NAMES[r["track"]]["layouts"][r["path"]]
        # track = location + layout
        self.track = f"{self.location} ({self.layout})"

        rc = r["raceConfiguration"]

        if rc["proximityGhostingMode"] == 1:
            self.collisions = "Prevent All Collisions"
        elif rc["proximityGhostingMode"] == 2:
            self.collisions = "Prevent Lapping Collisions"
        else:
            self.collisions = "All Collisions"

        self.has_wear = rc["wearMode"]
        # self.tire_wear = int(rc["wearCoefficients"]["tireWear"] * 100)
        # self.fuel_use = int(rc["wearCoefficients"]["fuelUse"] * 100)
        # self.vehicle_damage = int(rc["wearCoefficients"]["vehicleDamage"] * 100)
        # self.env_damage = int(rc["wearCoefficients"]["environmentDamage"] * 100)
        self.slipstream = int(rc["geometricSlipStreamCoefficient"] * 100)
        self.rubberbanding = int(rc["standingsSlipStreamCoefficient"] * 100)

        rec = rc["raceEndConditionConfig"]

        if rec["type"] == 1:
            self.length_type = "lap"
            self.length = rec["parameter"]
        elif rec["type"] == 2:
            self.length_type = "minute"
            self.length = rec["parameter"] / 60
        else:
            self.length_type = "km"
            self.length = rec["parameter"] / 1000

        self.length_desc = f"{self.length} {self.length_type}s"

    @property
    def datetime_race(self):
        return self.datetime_utc


class Race(Session):
    def __init__(self, race_file_content: dict) -> None:
        super().__init__(race_file_content)

        self.results = [
            RaceResult(result_dict)
            for result_dict in race_file_content["raceResult"]["racerResults"]
        ]


class Quali(Session):
    def __init__(self, race_file_content: dict) -> None:
        super().__init__(race_file_content)

        self.results = [
            QualiResult(result_dict)
            for result_dict in race_file_content["raceResult"]["racerResults"]
        ]



from pathlib import Path
from typing import List
from ksup_analyzer.replay.ReplayFile import QualiFile, RaceFile


class FileHandler:
    def __init__(self, file_paths: List[Path]) -> None:
        self.file_paths = [Path(file_path) for file_path in file_paths]
        self._quali_file, self._race_file = self.assign_files(self.file_paths)

    def get_quali_file_content(self):
        return self._quali_file.content

    def get_race_file_content(self):
        return self._race_file.content

    def assign_files(self, file_paths: List[Path]):
        self.check_input_files(file_paths)

        quali_file = None
        race_file = None

        for file_path in file_paths:
            with open(file_path, "r") as file:
                file_content = file.read()

                if (
                    '\\"text\\":\\"Qualifying\\"' in file_content
                    or '\\"text\\":\\"Qualifier\\"' in file_content
                ):
                    quali_file = QualiFile(self, file_content)
                elif '\\"text\\":\\"Race\\"' in file_content:
                    race_file = RaceFile(self, file_content)
                else:
                    raise Exception("Did not find keywords in the files ")

        assert (
            race_file is not None
        ), "Did not find a race file! Check your input files!"

        return quali_file, race_file

    def check_input_files(self, file_paths: List[Path]):
        assert len(file_paths) in [
            1,
            2,
        ], "You can pass up to two files. One (optional) qualifying file and one mandatory race replay .header file."

        assert all(
            file_path.suffix == ".header" for file_path in file_paths
        ), "The file(s) you provided are not .header files!"

        if len(file_paths) == 2:
            assert file_paths[0] != file_paths[1], "You provided the same file twice!"



import numpy as np
import pandas as pd

pd.set_option("mode.chained_assignment", None)


class Result:
    def __init__(self, results_dict: dict) -> None:
        self.assign_individual_properties(results_dict)
        self.series = None

    def assign_individual_properties(self, res: dict) -> None:
        """
        When it comes to results, there are some that describe the drivers individual performance like lap times
        and some that need more context, namely times of other drivers in order to infer positions for example.

        This method uses the (mostly) individual results given in the replay files. Only exception is the "num_laps_led"
        property which includes knowledge of other drivers times of course.
        """
        self.driver_id = res["racingTeamId"]

        self.starting_position = res["startPositionIndex"] + 1
        # this is the total time from session begin until the driver finsished the last lap
        # this is None if the driver did not finish
        # it does include the time a driver needs in "lap 0" until the beginning of lap 1 (crossing the start/finish line)
        self.total_time = res["finishTime"]
        # list of lap times
        self.lap_times = res["lapTimes"]
        # number of laps driven, relevant for lappings
        self.num_laps_driven = len(self.lap_times) if self.lap_times else np.nan
        # minimum lap time
        self.fastest_lap_time = min(self.lap_times) if self.lap_times else np.nan
        # list of time penalties per lap, it does not include hitting CCDs during the race cause that
        # only slows you down (it does not give a time penalty)
        self.lap_time_penalties = res["lapTimePenalties"]
        # how many laps the driver led; interesting in a multi-lap quali
        self.num_laps_led = res["numLapsLed"]
        # metres driven, not sure if relevant
        self.metres_driven = res["metresDriven"]

        # as the total time includes the time from the starting position to the start/finish line
        # and the lap times do not, we can extract the time the driver needed to get there
        # This is relevant to calculate positions per lap because we need to add this time to the lap times
        # in order to calculate positions
        self.time_until_starting_line = (
            self.total_time - sum(self.lap_times) if self.total_time else None
        )

    def as_series(self) -> pd.Series:
        if self.series is None:
            property_attributes = [
                attr
                for attr in dir(self)
                if not attr.startswith("__")
                and not callable(getattr(self, attr))
                and attr not in ["series", "driver_id"]
            ]

            indices = property_attributes

            data = [getattr(self, attr) for attr in property_attributes]

            self.series = pd.Series(data=data, index=indices, name=self.driver_id)

        return self.series


class RaceResult(Result):
    def __init__(self, results_dict: dict) -> None:
        super().__init__(results_dict)


class QualiResult(Result):
    def __init__(self, results_dict: dict) -> None:
        super().__init__(results_dict)


class RaceResultsDataFrame(pd.DataFrame):
    # this makes sure that when you would normally construct a new dataframe when applying pandas functions
    # it instead will construct a RaceResultsDataFrame instance making sure that the methods defined below
    # are still available
    @property
    def _constructor(self):
        return RaceResultsDataFrame

    @property
    def has_quali_data(self):
        return any([col.endswith("_quali") for col in self.columns])

    @property
    def participants(self):
        return len(self.index)

    @property
    def lap_position_table(self):
        if "lap_positions_race" not in self.columns:
            self.__calc_lap_positions()
            print(
                "WARNING: You tried to get the lap_position_table before running '__calc_lap_positions()'. I did this for you although you should handle that. Do not run it twice."
            )

        # create lap table dataframe that has a column for every lap and a row for every driver
        df = pd.DataFrame(self["lap_positions_race"].to_list())
        df.index = self["name"]

        # transposed every column name is a driver name and every row index is the laps completed
        df = df.transpose()
        df.index.rename("laps completed", inplace=True)

        return df

    @property
    def gap_to_winner_table(self):
        if "gap_to_winner_race" not in self.columns:
            self.__calc_gap_to_winner()
            print(
                "WARNING: You tried to get the gap_to_winner_table before running '__calc_gap_to_winner()'. I did this for you although you should handle that. Do not run it twice."
            )

        # create lap table dataframe that has a column for every lap and a row for every driver
        df = pd.DataFrame(self["gap_to_winner_race"].to_list())
        df.index = self["name"]

        # transposed every column name is a driver name and every row index is the laps completed
        df = df.transpose()
        df.index.rename("laps completed", inplace=True)

        return df

    @property
    def gap_to_leader_table(self):
        if "gap_to_leader_race" not in self.columns:
            self.__calc_gap_to_leader()
            print(
                "WARNING: You tried to get the gap_to_leader_table before running '__calc_gap_to_leader()'. I did this for you although you should handle that. Do not run it twice."
            )

        # create lap table dataframe that has a column for every lap and a row for every driver
        df = pd.DataFrame(self["gap_to_leader_race"].to_list())
        df.index = self["name"]

        # transposed every column name is a driver name and every row index is the laps completed
        df = df.transpose()
        df.index.rename("laps completed", inplace=True)

        return df

    def _run_result_calculations(self) -> None:
        """
        Until this point self is a pandas dataframe with all the information from the header file(s).

        These are the columns which are available (the _quali columns only if a quali header files was provided):
        - 'starting_position_race'
        - 'fastest_lap_time_race'
        - 'lap_time_penalties_race'
        - 'lap_times_race'
        - 'metres_driven_race'
        - 'num_laps_driven_race'
        - 'num_laps_led_race'
        - 'time_until_starting_line_race'
        - 'total_time_race'

        - 'fastest_lap_time_quali'
        - 'lap_time_penalties_quali'
        - 'lap_times_quali'
        - 'metres_driven_quali'
        - 'num_laps_driven_quali'
        - 'num_laps_led_quali'
        - 'time_until_starting_line_quali'
        - 'total_time_quali'

        - 'car'
        - 'colors'
        - 'is_ai'
        - 'name'
        - 'platform'
        - 'vehicle_colors'

        Now it is about calculating more valuable variables out of the data available, f.e. positions per lap.
        """
        self.__calc_race_position()
        self.__interpolate_time_until_starting_line_race()
        self.__calc_lap_positions()
        self.__calc_gap_to_winner()
        self.__calc_gap_to_leader()

    def __calc_race_position(self):
        self.sort_values(
            by=["num_laps_driven_race", "total_time_race", "starting_position_race"],
            ascending=[False, True, True],
            inplace=True,
        )
        self["end_position_race"] = list(range(1, len(self.index) + 1))

    def __interpolate_time_until_starting_line_race(self):
        # creating a series to create an index with equal distances for interpolation
        time_until_starting_line_race = self["time_until_starting_line_race"]
        time_until_starting_line_race.index = list(range(1, len(self.index) + 1))

        # spline is able to extrapolate as well
        # order 2 as the speed difference drivers have when crossing the start finish line decreases
        # the further back you go in the starting grid (P1 much slower than P2 but P11 and P12 similar if not identical)
        self[
            "time_until_starting_line_race"
        ] = time_until_starting_line_race.interpolate(
            method="spline", order=2, inplace=False
        ).values

    def __calc_lap_positions(self):
        """
        We want to calculate the position of each car on each lap and add a list of positions as a new column in self.
        In order reduce complexity we will create another dataframe called df that only holds the information we need.
        """

        # remove those that have no lap times and reduce the columns to what we really need
        df = self.loc[
            ~self["lap_times_race"].isna(),
            [
                "name",
                "time_until_starting_line_race",
                "lap_times_race",
                "starting_position_race",
                "end_position_race",
            ],
        ]

        df_lap_times = pd.DataFrame(df["lap_times_race"].to_list())
        df_lap_times.index = df.index

        df_lap_times = pd.concat(
            [df["time_until_starting_line_race"], df_lap_times],
            axis=1,
            ignore_index=True,
        )
        df_lap_times = df_lap_times.cumsum(axis=1)
        df = pd.concat([df, df_lap_times], axis=1)
        del df["lap_times_race"]

        # now we have a df with the team id as an index and the lap numbers (from 0 to e.g. 10) as a column
        # we do know the starting positions and the finishing positions already via the "starting_position" / "race_position" columns within df

        for col in df_lap_times.columns:
            df.sort_values(by=col, inplace=True)
            df[f"Lap {col}"] = list(range(1, len(df.index) + 1))
            df.loc[df[col].isnull(), f"Lap {col}"] = np.NaN
            col_pos_last_lap = f"Lap {col}"

        for _, row in df[["name", col_pos_last_lap, "end_position_race"]].iterrows():
            if np.isnan(row[col_pos_last_lap]):
                continue

            if row[col_pos_last_lap] != row["end_position_race"]:
                print(
                    f'Calculated position seems to be incorrect. Expected pos {row["end_position_race"]} at the end for {row["name"]} but found {row[col_pos_last_lap]}.'
                )

        df.sort_values(by="end_position_race", inplace=True)

        df[col_pos_last_lap] = df["end_position_race"]

        df_lap_table = df.loc[
            :,
            ["starting_position_race"]
            + [
                col
                for col in df.columns
                if isinstance(col, str) and col.startswith("Lap") and col != "Lap 0"
            ],
        ]

        # if someone has been lapped we want to see the final position anyways
        # example: there are 20 laps and someone did 18. We know his position until lap 18 and in the finish (lap 20)
        # therefore when backfilling we fill lap 19 position with the position at lap 20 cause that should be the same
        df_lap_table = df_lap_table.bfill(axis=1)

        # add the lap positions in a column of self as a list
        df_lap_table["lap_positions_race"] = df_lap_table.values.tolist()

        self["lap_positions_race"] = np.nan

        for i, value in df_lap_table["lap_positions_race"].items():
            # simply assigning a list does not work unfortunately:
            # error: ValueError: Must have equal len keys and value when setting with an iterable
            self.loc[self.index == i, "lap_positions_race"] = self.loc[
                self.index == i, "lap_positions_race"
            ].apply(lambda _: value)

    def __calc_gap_to_winner(self):
        """
        We want to calculate the gap to the winner of each car on each lap and add a list of gaps as a new column in self.
        In order reduce complexity we will create another dataframe called df that only holds the information we need.
        """

        # remove those that have no lap times and reduce the columns to what we really need
        df = self.loc[
            ~self["lap_times_race"].isna(),
            [
                "name",
                "time_until_starting_line_race",
                "lap_times_race",
                "starting_position_race",
                "end_position_race",
            ],
        ]

        df_lap_times = pd.DataFrame(df["lap_times_race"].to_list())
        df_lap_times.index = df.index

        df_lap_times = pd.concat(
            [df["time_until_starting_line_race"], df_lap_times],
            axis=1,
            ignore_index=True,
        )
        df_lap_times = df_lap_times.cumsum(axis=1)
        df = pd.concat([df, df_lap_times], axis=1)
        del df["lap_times_race"]

        # now we have a df with the team id as an index and the lap numbers (from 0 to e.g. 10) as a column
        # we do know the starting positions and the finishing positions already via the "starting_position" / "race_position" columns within df

        for col in df_lap_times.columns:
            df[f"Lap {col}"] = 0
            for idx in range(len(df.index)):
                df[f"Lap {col}"][idx] = df[col][idx] - df[col][0]
            df.loc[df[col].isnull(), f"Lap {col}"] = np.NaN

        df_gap_table = df.loc[
            :,
            [
                col
                for col in df.columns
                if isinstance(col, str) and col.startswith("Lap")
            ],
        ]

        df_gap_table = df_gap_table.bfill(axis=1)

        # add the gaps in a column of self as a list
        df_gap_table["gap_to_winner_race"] = df_gap_table.values.tolist()

        self["gap_to_winner_race"] = np.nan

        for i, value in df_gap_table["gap_to_winner_race"].items():
            # simply assigning a list does not work unfortunately:
            # error: ValueError: Must have equal len keys and value when setting with an iterable
            self.loc[self.index == i, "gap_to_winner_race"] = self.loc[
                self.index == i, "gap_to_winner_race"
            ].apply(lambda _: value)

    def __calc_gap_to_leader(self):
        """
        We want to calculate the gap to the leader of each car on each lap and add a list of gaps as a new column in self.
        In order reduce complexity we will create another dataframe called df that only holds the information we need.
        """

        # remove those that have no lap times and reduce the columns to what we really need
        df = self.loc[
            ~self["lap_times_race"].isna(),
            [
                "name",
                "time_until_starting_line_race",
                "lap_times_race",
                "starting_position_race",
                "end_position_race",
            ],
        ]

        df_lap_times = pd.DataFrame(df["lap_times_race"].to_list())
        df_lap_times.index = df.index

        df_lap_times = pd.concat(
            [df["time_until_starting_line_race"], df_lap_times],
            axis=1,
            ignore_index=True,
        )
        df_lap_times = df_lap_times.cumsum(axis=1)
        df = pd.concat([df, df_lap_times], axis=1)
        del df["lap_times_race"]

        # now we have a df with the team id as an index and the lap numbers (from 0 to e.g. 10) as a column
        # we do know the starting positions and the finishing positions already via the "starting_position" / "race_position" columns within df

        for col in df_lap_times.columns:
            df.sort_values(by=col, inplace=True)
            df[f"Lap {col}"] = 0.0
            for idx in range(len(df.index)):
                df.at[idx, f"Lap {col}"] = df[col].iloc[idx] - df[col].iloc[0]
            df.loc[df[col].isnull(), f"Lap {col}"] = np.NaN

        df.sort_values(by="end_position_race", inplace=True)

        df_gap_table = df.loc[
            :,
            [
                col
                for col in df.columns
                if isinstance(col, str) and col.startswith("Lap")
            ],
        ]

        df_gap_table = df_gap_table.bfill(axis=1)

        # add the gap in a column of self as a list
        df_gap_table["gap_to_leader_race"] = df_gap_table.values.tolist()

        self["gap_to_leader_race"] = np.nan

        for i, value in df_gap_table["gap_to_leader_race"].items():
            # simply assigning a list does not work unfortunately:
            # error: ValueError: Must have equal len keys and value when setting with an iterable
            self.loc[self.index == i, "gap_to_leader_race"] = self.loc[
                self.index == i, "gap_to_leader_race"
            ].apply(lambda _: value)



import json


class ReplayFile:
    def __init__(self, file_handler, file_content_raw: str) -> None:
        self.file_handler = file_handler
        self.content = self.read_content(file_content_raw)

    def read_content(self, file_content_raw: str) -> dict:
        file_content_str = self.convert_to_valid_json(file_content_raw)
        file_content_json = json.loads(file_content_str)
        self.double_check_json_content(file_content_json)
        return file_content_json

    def convert_to_valid_json(self, file_content_raw: str) -> str:
        """
        The content of the json files is not 100% valid json format.
        So to read it with json.loads() we first need to get rid of some backslashes and
        some double quotes around the raceConfiguration.
        """
        return (
            file_content_raw.replace("\\", "")
            .replace('"raceConfiguration":"', '"raceConfiguration":')
            .replace('"raceConfiguration": "', '"raceConfiguration": ')
            .replace('}","track":', '},"track":')
        )


class QualiFile(ReplayFile):
    def __init__(self, file_handler, file_content_raw: str) -> None:
        super().__init__(file_handler, file_content_raw)

    def double_check_json_content(self, file_content_json: dict) -> None:
        assert file_content_json["raceConfiguration"]["eventName"]["text"] in [
            "Qualifying",
            "Qualifier",
        ], "This does not seem to be a quali file!"


class RaceFile(ReplayFile):
    def __init__(self, file_handler, file_content_raw: str) -> None:
        super().__init__(file_handler, file_content_raw)

    def double_check_json_content(self, file_content_json: dict) -> None:
        assert file_content_json["raceConfiguration"]["eventName"]["text"] in [
            "Race"
        ], "This does not seem to be a race file!"



from ksup_analyzer.event.Driver import Driver


class LineUp:
    def __init__(self, race_file_content: dict) -> None:
        self.assign_properties(race_file_content)

    def assign_properties(self, r: dict) -> None:
        self.drivers = [
            Driver(did, driver_dict) for did, driver_dict in r["configsById"].items()
        ]

    def __str__(self) -> str:
        return ",".join([str(driver) for driver in self.drivers])
